source('modules/Classes/Transition.R')
source('modules/Classes/PayoffFunction.R')

DPEngine <- 
  setRefClass(
    'DPEngine',
    fields = list(
      beta = 'numeric',
      tran = 'Transition',
      payoff = 'PayoffFunction',
      CCP = 'list',
      choice.specific.value = 'list',
      value.fn = 'numeric'
    ),
    methods = list(
      initializeDP = function(theta, dirichlet.alpha, num.states, num.actions){
        tran <<-
          Transition$new(
            num.states = num.states,
            num.actions = num.actions,
            dirichlet.alpha = dirichlet.alpha
          ) %T>% {
            .$calc_transition()
          }
        
        payoff <<-
          PayoffFunction$new(theta = theta) %T>% {
            .$createPayoffStats(num.states, num.actions)
            .$updatePayoffs()
          }
        
        value.fn <<- rep(0, num.states)
        
        solve_RVI()
      },
      
      updateChoiceSpecificValue = function() {
        #Maps payoff$utility and value.fn to choice.specific.value
        choice.specific.value <<-
          map2(
            payoff$utility,
            tran$tran.mat,
            ~ (.x + beta * .y %*% value.fn) %>% 
              as.numeric
          )
      },
      
      updateCCPs = function(){
        #Maps value.fn to CCP
        updateChoiceSpecificValue()
        
        max.vals <- 
          choice.specific.value %>% 
          rd(pmax)
        
        CCP <<-
          choice.specific.value %>% 
          map(~exp(. - max.vals)) %>% {
            denom <- rd(., `+`)
            map(., ~ c(./denom))
          } %>%
          map(~ pmax(10^-100, .))
      },
      
      operator_VI = function(){
        #Maps value.fn to value.fn
        updateChoiceSpecificValue()
        
        max.vals <- 
          choice.specific.value %>% 
          rd(pmax)
        
        value.fn <<-
          choice.specific.value %>% 
          map(~exp(. - max.vals)) %>% 
          rd(`+`) %>% 
          log +
          max.vals
      },
      
      solve_RVI = function(delta = 10^-8){
        dif <- Inf
        while(dif > delta){
          value.fn.l <- value.fn
          operator_VI()
          value.fn <<- value.fn - value.fn[1]
          dif <- max(abs(value.fn.l - value.fn))
        }
        updateCCPs()
      }
    )
  )