source('modules/Classes/DPEngine.R')

Estimator <- 
  setRefClass(
    'Estimator',
    fields = list(
      beta = 'numeric',
      num.states = 'numeric',
      num.actions = 'numeric',
      num.theta = 'numeric',
      sim.horizons = 'list',
      theta = 'numeric',
      theta.estimate = 'numeric',
      dp = 'DPEngine'
    ),
    methods = list(
      initializeEstimator = function(){
        beta <<- .99
        num.states <<- 100
        num.theta <<- 2
        num.actions <<- 3
        dirichlet.alpha <- 1
        
        theta <<- rnorm(num.theta)
        
        dp <<- 
          DPEngine$new(beta = beta) %T>% {
            .$initializeDP(
              theta = theta,
              dirichlet.alpha = dirichlet.alpha,
              num.states = num.states, 
              num.actions = num.actions
            )
          }
        
        calcSimHorizons()
      },
      
      estimateModel = function(num.paths, num.periods) {
        sim.data <-
          seq(num.actions) %>% 
          map(function(initial.action){
            stat <- rep(0, num.theta)
            error <- 0
            for(s in seq(num.paths)){
              my.state <- 1
              my.action <- initial.action
              for(t in (seq(num.periods) - 1)){
                stat <-  stat + beta^t * dp$payoff$payoff.stat[[my.action]][my.state,]
                error <- error - (t>0) * beta^t * log(dp$CCP[[my.action]][my.state])
                
                my.state <- 
                  sample(
                    x = seq(num.states), 
                    size = 1,
                    prob = dp$tran$tran.mat[[my.action]][my.state,]
                  )
                
                my.action <- 
                  dp$CCP %>% 
                  transpose %>% {
                    .[[my.state]]
                  } %>% 
                  un %>% 
                  sample(
                    x = seq(num.actions),
                    size = 1,
                    prob = .
                  )
              }
            }
            list(
              stat = stat / num.paths,
              error = error / num.paths
            )
          })
        
        val.diff <-
          dp$CCP %>% 
          transpose %>% {
            .[[1]]
          } %>% 
          un %>% 
          df %>% 
          mt(val.diff = log(.) - first(log(.))) %>% 
          sl(val.diff) %>% 
          slice(-1) %>% 
          un
        
        est.diff <-
          sim.data %>% 
          transpose %>% 
          map(~{rd(., rbind)}) %>% 
          map(~{
            t(t(.) - .[1, ]) %>% {
              .[-1,]
            }
          })
        
        theta.estimate <<- c(solve(t(est.diff$stat) %*% est.diff$stat, t(est.diff$stat)%*% (val.diff - est.diff$error)))
      },
      
      calcSimHorizons = function(epsilon = 10^-6) {
        sim.horizons <<-
          list(
            regular.convergence = floor(log(epsilon)/log(beta)) + 1,
            strong.convergence = 
              map2(dp$tran$tran.mat, dp$CCP, ~{.x * .y}) %>% 
              rd(`+`) %>% 
              eigen %>% {
                .$values
              } %>% 
              Mod %>% {
                .[2]
              } %>% {
                floor(log(epsilon)/log(beta * .)) + 1
              }
          )
      }
    )
  )