PayoffFunction <- 
  setRefClass(
    'PayoffFunction',
    fields = list(
      theta = 'list',
      profit = 'data.frame'
    ),
    methods = list(
      calc_value = function(omega){
        ifelse(
          omega <= 5, 
          3 * omega - 4,
          12 + log(2 - exp(16 - 3 * omega))
        )
      },
      
      calc_quantity = function(price, omega){
        (calc_value(omega) - price) %>% {
          exp(. - max(.))/(exp(-max(.)) + sum(exp(. - max(.))))
        } %>% {
          theta$m * .
        }
      },
      
      calc_profit = function(){
        profit_fn <- function(price, omega){
          calc_quantity(price, omega) * (price - theta$c)
        }
        
        price <-
          expand.grid(
            omega.i = seq(theta$M),
            omega.j = seq(theta$M)
          ) %>% 
          br(
            function(my.omega){
              delta <- 10^-10
              dif <- Inf
              price <- rep(1, 2)
              while(dif > delta){
                dif <- 0
                for(i in 1:2){
                  price.l <- price[i]
                  
                  agent_i_problem <- function(price.i){
                    price %>% {
                      .[i] <- price.i
                      .
                    } %>% 
                      profit_fn(
                        price = ., 
                        omega = un(my.omega)
                      ) %>% {
                        .[i]
                      }
                  }
                  
                  price[i] <- 
                    optimize(
                      f = agent_i_problem,
                      lower = theta$c,
                      upper = 25,
                      maximum = TRUE
                    ) %>% {
                      .$maximum
                    }
                  
                  dif <- dif + abs(price[i] - price.l)
                }
              }
              price
            },
            .collate = 'cols'
          ) %>% 
          rn(
            price.i = .out1,
            price.j = .out2
          )
        
        profit <<-
          price %>% 
          br(
            ~{
              profit_fn(
                price = c(.$price.i, .$price.j), 
                omega = c(.$omega.i, .$omega.j)
              ) %>% {
                .[1]
              }
            },
            .collate = 'cols'
          ) %>% 
          rn(profit = .out) %>% 
          sl(omega.i, omega.j, profit)
      }
    )
  )
    
    
    
    
    
    
    
    
    
    
    
    
    